package de.fraunhofer.iem.dataprovider.taskManager

import de.fraunhofer.iem.dataprovider.gitlab.GitConfiguration
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.taskManager.tasks.*
import jakarta.annotation.PreDestroy
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ObsoleteCoroutinesApi
import kotlinx.coroutines.channels.SendChannel
import kotlinx.coroutines.channels.actor
import org.springframework.stereotype.Component
import java.util.*


sealed class Event
class RepoChangedEvent(val gitConfiguration: GitConfiguration, val repoId: Long) : Event()
class SequenceTaskDoneEvent(val sequenceId: UUID) : Event()

/**
 * The Task Manager takes tasks and distributes them to
 * underlying workers. Internally it uses a channel
 * to manage incoming tasks.
 */
@Component
class TaskManager {

    // The used default dispatcher is ok for CPU-bound workloads. However,
    // if they block for a long time it's better to use a custom thread
    // pool solution.
    private val worker = Worker("Worker")

    // Should be used for long-running tasks, which DON'T use CPU power and
    // are non-blocking
    private val ioWorker = Worker("IO-Worker", CoroutineScope(Dispatchers.IO))
    // Should be used for long-running CPU heavy tasks, which potentially block.
    // private val customWorker = getCustomThreadpoolWorker()

    private val mainScope = CoroutineScope(Dispatchers.Default)
    private val actor: SendChannel<Event> = mainScope.taskActor()

    suspend fun addEvent(event: Event) {
        actor.send(event)
    }

    @OptIn(ObsoleteCoroutinesApi::class)  // This API is meant to be improved since 2017... so this should be good to use.
    // IntelliJ might tell u that it is ok to remove <ITask> from the actor declaration.
    // IT IS NOT! this will lead to very strange gradle compile errors!
    private fun CoroutineScope.taskActor() = actor<Event> {
        with(MyActor()) {
            for (task in channel) {
                onReceive(task)
            }
        }
    }

    @PreDestroy
    fun close() {
        actor.close()
        worker.close()
        ioWorker.close()
    }

    /**
     * The state in this internal class is thread safe.
     * See https://kotlinlang.org/docs/shared-mutable-state-and-concurrency.html#actors
     * It is used to manage e.g. fixed task sequences, as well as react to general events.
     */
    private inner class MyActor {

        private val logger = getLogger(javaClass)
        private val taskSequences = HashMap<UUID, TaskSequence>()
        suspend fun onReceive(event: Event) {
            logger.info("[${Thread.currentThread().name}] add task called in Task Manager $event")

            when (event) {
                is RepoChangedEvent -> {
                    val taskSequence = TaskSequence.getRepoChangedSequence(
                        event.repoId,
                        event.gitConfiguration,
                        responseChannel = ::addEvent
                    )
                    taskSequences[taskSequence.sequenceID] = taskSequence
                    taskSequence.next().forEach {
                        worker.addTask(it)
                    }
                }

                is SequenceTaskDoneEvent -> handleSequenceTaskDone(event.sequenceId)
            }
            logger.info("[${Thread.currentThread().name}] add task finished in Task Manager $event")
        }

        private suspend fun handleSequenceTaskDone(id: UUID) {

            val sequence = taskSequences.getOrDefault(id, TaskSequence.EmptySequence)

            if (sequence.hasNext()) {
                sequence.next().forEach {
                    when (it) {
                        is GetGitlabProjectTask -> ioWorker.addTask(it)
                        else -> worker.addTask(it)
                    }
                }
            } else {
                taskSequences.remove(id)
                logger.info("[${Thread.currentThread().name}] sequence with id $id finished.")
            }
        }

    }
}

private class TaskSequence(
    private val tasks: List<List<ITask>>,
    val sequenceID: UUID = UUID.randomUUID()
) : Iterator<List<ITask>> {

    companion object Factory {
        val EmptySequence = TaskSequence(listOf())
        fun getRepoChangedSequence(
            repoId: Long,
            gitConfiguration: GitConfiguration,
            responseChannel: suspend (task: Event) -> Unit
        ): TaskSequence {
            val id = UUID.randomUUID()
            val gitlabTask = GetGitlabProjectIdTask(repoId, gitConfiguration, responseChannel, id)
            val gitTask = GitTask(TaskType.REPO_CHANGED, gitlabTask::getResult, responseChannel, id)
            return TaskSequence(
                listOf(
                    listOf(gitlabTask),
                    listOf(gitTask),
                    listOf(
                        JavaTask(TaskType.REPO_CHANGED, "-version", responseChannel = responseChannel, sequenceId = id),
                        JavaTask(TaskType.REPO_CHANGED, "-help", responseChannel = responseChannel, sequenceId = id)
                    )
                ),
                id
            )
        }
    }

    private val it = tasks.iterator()

    override fun next(): List<ITask> {
        return it.next()
    }

    override fun hasNext(): Boolean {
        return it.hasNext()
    }
}
